local floor, ceil, max, min = math.floor, math.ceil, math.max, math.min

--------------------------------------------------------------------------------
-- Sprite Entities
local Entity, Collectible, Projectile, Money, Enemy, Player

--------------------------------------------------------------------------------
do
  Entity = class {
    type = "Entity",
    x = 0, y = 0, w = 1, h = 1, dx = 0, dy = 0,
    facing = "right", bank_name = "default", frame = "default",
    jump_height = 260, speed = 112,
    gravity = 720, terminal_velocity = 256,
    grounded = false, jumping = false, tangible = true, hurts = false,

    handleCollision = NULLFUNC,
    pickup = NULLFUNC,
  }

  function Entity:init(parent, x, y)
    self.parent = parent
    self.x = x or self.x
    self.y = y or self.y
  end

  function Entity:applyPhysics(dt)
    local x, y = self.x, self.y
    local dx, dy = self.dx, self.dy

    x = x + dx*dt

    dy = min(dy + self.gravity*(dt*0.5), self.terminal_velocity)
    y = y + dy*dt
    dy = min(dy + self.gravity*(dt*0.5), self.terminal_velocity)

    self.x, self.y, self.dx, self.dy, self.grounded = self.parent:resolveCollisions(self, x, y, dx, dy)
    if self.jumping and self.dy >= 0 then self.jumping = false end
  end

  function Entity:update(dt)
    self:applyPhysics(dt)
  end

  function Entity:draw(dt, ox, oy)
    local x, y = self.x+self.w*0.5, self.y+self.h-1
    Graphics:setColor(Graphics.Color.White)
      :drawSprite(self.bank_name, self.frame, x-ox, y-oy, (self.facing=="left"))
  end

  function Entity:overlaps(other)
    return util.rectOverlaps(self.x, self.y, self.w, self.h, other.x, other.y, other.w, other.h)
  end
end

--------------------------------------------------------------------------------
do
  Projectile = Entity:subclass {
    bank_name = "icons",
    speed = 180,
    gravity = 0,
  }

  function Projectile:init(parent, x, y, dx, dy, lifetime)
    Projectile.super.init(self, parent, x, y)
    self.clock, self.lifetime = 0, lifetime
    self.dx, self.dy = dx * self.speed, dy * self.speed
    self.animator = Animator(self.animations)
    self.animator:reset("default")
  end

  function Projectile:update(dt)
    Projectile.super.update(self, dt)
    self.clock = self.clock + dt
    if (self.dx==0 and self.dy==0) or (self.clock >= self.lifetime) then
      self.parent:removeEntity(self)
    end
  end

  function Projectile:draw(dt, ox, oy)
    self.animator:update(dt)
    self.frame = self.animator:get()
    Projectile.super.draw(self, dt, ox, oy)
  end

  ThrowingKnife = Projectile:subclass {
    animations = {
      default = {
        { "knife_1", 0.1 },
        { "knife_2", 0.1 },
        { "knife_3", 0.1 },
        { "knife_4", 0.1 },
      }
    }
  }
end

--------------------------------------------------------------------------------
do
  Collectible = Entity:subclass {
    type = "Collectible",
    bank_name = "icons",
    pickedUp = NULLFUNC
  }

  function Collectible:handleCollision(other)
    if other:pickup(self) then
      self:pickedUp()
      self.parent:removeEntity(self)
    end
  end

  Money = Collectible:subclass {
    frame = "money",
  }

end

--------------------------------------------------------------------------------
do
  Enemy = Entity:subclass {
    type = "Enemy",
    w = 15, h = 15,
    bank_name = "player",

  }

  function Enemy:init(parent, x, y)
    Enemy.super.init(self, parent, x, y)
    self.animator = Animator(Player.animations)
    self.animator:reset("standing")
  end

  function Enemy:update(dt)
    


    Enemy.super.update(self, dt)
  end

  function Enemy:draw(dt, ox, oy)
    self.animator:update(dt)
    self.frame = self.animator:get()
    Enemy.super.draw(self, dt, ox, oy)
  end

end

--------------------------------------------------------------------------------
do
  Player = Entity:subclass {
    type = "Player",
    w = 13, h = 15,
    bank_name = "player",

    pickup = function() return true end,

    sprite_bounds = {
      stand = { x =   0, y =  0, w = 24, h = 24, u =  12, v = 23 },
      blink = { x =  24, y =  0, w = 24, h = 24, u =  36, v = 23 },
      run_1 = { x =  72, y =  0, w = 24, h = 24, u =  84, v = 23 },
      run_2 = { x =  96, y =  0, w = 16, h = 24, u = 105, v = 23 },
      run_3 = { x = 112, y =  0, w = 24, h = 24, u = 128, v = 23 },
      shoot = { x = 136, y =  0, w = 32, h = 24, u = 148, v = 23 },
      jump  = { x =   0, y = 24, w = 24, h = 32, u =  12, v = 48 },
      hurt  = { x =  24, y = 24, w = 24, h = 32, u =  36, v = 48 },
      jumpshoot = { x = 48, y = 24, w = 32, h = 32, u = 60, v = 48 },
    },

    animations = {
      standing = {
        { "stand", 1.75 },
        { "blink", 0.25 },
      },
      running = {
        { "run_1", 0.17 },
        { "run_2", 0.17 },
        { "run_3", 0.17 },
        { "run_2", 0.17 },
      },
      shooting = {
        { "shoot", Animator.FREEZE },
      },
      jumping = {
        { "jump", Animator.FREEZE },
      },
      jumpingshoot = {
        { "jumpshoot", Animator.FREEZE },
      },
      hurt = {
        { "hurt", Animator.FREEZE },
      },
    }
  }

  function Player:init(parent, x, y)
    Player.super.init(self, parent, x, y)
    self.animator = Animator(self.animations)
    self.animator:reset("standing")
  end

  function Player:update(dt)
    if Input.hold.left then
      self.facing = "left"
      self.dx = -self.speed
      if self.grounded then
        self.animator:set("running")
      end
    elseif Input.hold.right then
      self.facing = "right"
      self.dx = self.speed
      if self.grounded then
        self.animator:set("running")
      end
    else
      self.dx = 0
      if self.grounded then
        self.animator:set("standing")
      end
    end

    if self.jumping and not Input.hold.jump then
      self.jumping = false
      self.dy = self.dy *0.70
    end

    if Input.hold.jump and Input.len.jump < 0.1 and self.grounded then
      self.animator:set("jumping")
      self.grounded = false
      self.jumping = true
      self.dy = -self.jump_height
    end

    if Input.tap.shoot then
      self.shooting = 0.0
      local dx = (self.facing == "left") and -1 or 1
      local e = ThrowingKnife(self.parent, self.x+(self.w*0.5), self.y+(self.h*0.5), dx, 0, 4)
      self.parent:addEntity(e)
    end

    if self.shooting then
      self.animator:set(self.grounded and "shooting" or "jumpingshoot")
      self.shooting = self.shooting + dt
      if self.shooting >= 0.16 then self.shooting = false end
    else
      if not self.grounded then
        self.animator:set("jumping")
      elseif self.dx == 0 then
        self.animator:set("standing")
      end
    end

    Player.super.update(self, dt)
  end

  function Player:draw(dt, ox, oy)
    self.animator:update(dt)
    self.frame = self.animator:get()
    Player.super.draw(self, dt, ox, oy)
  end
end

--------------------------------------------------------------------------------

local entity = {
  Entity = Entity,
  Collectible = Collectible,
  Projectile = Projectile,
  Money = Money,
  Enemy = Enemy,
  Player = Player,
}

function entity.init()
  Graphics:loadSpriteBank("player", "gfx/lbjsprite.png", Player.sprite_bounds)
end

return entity
