local floor, ceil, max, min = math.floor, math.ceil, math.max, math.min
local round = function(x) return floor(x+0.5) end

--------------------------------------------------------------------------------
-- Graphic utilities

local function C(r, g, b, a) return {r=r, g=g, b=b, a=a or 255} end

local Graphics = {
  game_width = 240,
  game_height = 160,
  scale = 1,
  max_scale = 3,
  sprites = {},
  images = {},

  Color = {
    Black     = C(  0,   0,   0),
    DarkGray  = C( 96,  96,  96),
    LightGray = C(192, 192, 192),
    White     = C(255, 255, 255),
    Red       = C(255,   0,   0),
    Green     = C(  0, 255,   0),
    Blue      = C(  0,   0, 255),
  },
  last_color = C(-1, -1, -1, -1)
}

function Graphics:init()
  self.scale = max(1,floor(love.graphics.getWidth()/self.game_width))
  self.max_scale = self:determineMaxScale()
  love.graphics.setColorMode("modulate")
  love.graphics.setBlendMode("alpha")
  love.mouse.setVisible(false)
  self:initPixels()

  -- self:loadSpriteData("ui", "data/ui_bounds.lua")
  self:loadFont("gfx/cgafont.png")

  self:loadImage("lbjtitle", "gfx/lbj-title.png")
  self:loadInvertedImage("lbjinverted", "gfx/lbj-title.png")
  return self
end

function Graphics:determineMaxScale()
  local w, h = 640, 480 
  local modes = love.graphics.getModes()
  for _, mode in pairs(modes) do
    w = max(w, mode.width)
    h = max(h, mode.height)
  end
  local x, y = floor(w/self.game_width), floor(h/self.game_height)
  return max(self.max_scale, min(x, y))
end

function Graphics:setNextScale()
  return self:changeScale((self.scale%self.max_scale)+1)
end

function Graphics:changeScale(size)
  self.scale = size
  love.graphics.setMode(self.game_width*size, self.game_height*size, false)
  love.mouse.setVisible(false)
  return self
end

function Graphics:saveScreenshot()
  local name = string.format("screenshot-%s.png", os.date("%Y%m%d-%H%M%S"))
  local shot = love.graphics.newScreenshot()
  shot:encode(name, "png")
  return self
end

function Graphics:initPixels()
  local pixelData = love.image.newImageData(4, 1)
  local c = { self.Color.Black, self.Color.DarkGray, self.Color.LightGray, self.Color.White }
  for i = 1, 4 do
    pixelData:setPixel(i-1, 0, c[i].r, c[i].g, c[i].b, 255)
  end
  self.pixelImage = love.graphics.newImage(pixelData)
  self.pixelImage:setFilter("nearest", "nearest")
  self.pixelQuads = {}
  for i = 1, 4 do
    self.pixelQuads[i] = love.graphics.newQuad(i-1, 0, 1, 1, 4, 1)
  end
  return self
end

function Graphics:drawPixel(x, y, c)
  local q = self.pixelQuads[c] or self.pixelQuads[1]
  love.graphics.drawq(self.pixelImage, q, round(x), round(y))
  return self
end

function Graphics:setColor(r, g, b, a)
  if type(r)=="table" then
    r, g, b, a = r.r or 255, r.g or 255, r.b or 255, r.a or 255
  else
    r, g, b, a = r or 255, g or 255, b or 255, a or 255
  end
  local last = self.last_color
  if (last.r~=r) or (last.g~=g) or (last.b~=b) or (last.a~=a) then
      last.r, last.g, last.b, last.a = r, g, b, a
      love.graphics.setColor(r, g, b, a)
  end
  return self
end

function Graphics:loadFont(name)
  local fontset = [=[ !"#$%&'()*+,-./0123456789:;<=>?]=] ..
    [=[@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_]=] ..
    [=[`abcdefghijklmnopqrstuvwxyz{|}~]=]
  self.fontimg = love.graphics.newImage(name)
  self.fontimg:setFilter("nearest", "nearest")
  self.font = love.graphics.newImageFont(self.fontimg, fontset)
  self.fontHeight = self.fontimg:getHeight()
  self.font:setLineHeight(self.fontHeight)
  love.graphics.setFont(self.font)
  return self
end

function Graphics:write(x, y, str, ...)
  str = str:format(...)
  if y == "center" then
    local lines = 0
    for _ in str:gmatch("[^\r\n]+") do lines = lines + 1 end
    y = floor((self.gameHeight-(self.fontHeight+2)*lines)*0.5)
  end
  for line in str:gmatch("[^\r\n]+") do
    local lx = (x == "center") and floor((self.gameWidth*0.5)-(line:len()*4)) or x
    love.graphics.print(line, lx, y)
    y = y + self.fontHeight + 2
  end
  return self
end

function Graphics:loadImage(name, filename)
  local image = love.graphics.newImage(filename)
  image:setFilter("nearest", "nearest")
  self.images[name] = image
  return self
end

local function inverterFunc(x, y, r, g, b, a)
  return 255-r, 255-g, 255-b, a
end

function Graphics:loadInvertedImage(name, filename)
  local data = love.image.newImageData(filename)
  data:mapPixel(inverterFunc)   
  local image = love.graphics.newImage(data)
  self.images[name] = image
  return self
end

function Graphics:drawImage(name, x, y)
  local image = self.images[name]
  if image then
    love.graphics.draw(image, x, y)
  end
  return self
end

function Graphics:loadSpriteBank(bankname, filename, bounds)
  if self.sprites[bankname] then
    -- already loaded, don't waste time loading again
    return self
  end

  local bank = {}
  bank.image = love.graphics.newImage(filename)
  bank.image:setFilter("nearest", "nearest")
  local sw, sh = bank.image:getWidth(), bank.image:getHeight()

  bank.rects = {}
  for k, v in pairs(bounds) do
    bank.rects[k] = {
      quad=love.graphics.newQuad(v.x, v.y, v.w, v.h, sw, sh),
      w = v.w, h = v.h,
      ox = (v.u) and (v.u - v.x) or 0,
      oy = (v.v) and (v.v - v.y) or 0,
    }
  end

  self.sprites[bankname] = bank
  return self
end

function Graphics:drawSprite(bankname, quadname, x, y, flipped)
  local bank = self.sprites[bankname]
  if not bank then return self end
  local rect = bank.rects[quadname]
  if not rect then return self end
  local horiz, ox
  if flipped then
    x, ox, horiz = x + rect.w, rect.w - rect.ox, -1
  else
    horiz, ox = 1, rect.ox
  end
  love.graphics.drawq(bank.image, rect.quad, x-ox, y-rect.oy, 0, horiz, 1)
  return self
end

function Graphics:loadTileset(name, filename, firstgid)
  self.imagecache = self.imagecache or {}
  if self.imagecache[name] then return end
  if self.image then return end
  local image = love.graphics.newImage(filename)
  image:setFilter("nearest", "nearest")
  local sw, sh = image:getWidth(), image:getHeight()
  self.quads = self.quads or {}
  self.tileimage = self.tileimage or {}
  local i=firstgid or 1
  for y = 0, sh-1, 16 do
    for x = 0, sw-1, 16 do
      self.quads[i] = love.graphics.newQuad(x, y, 16, 16, sw, sh)
      self.tileimage[i] = name
      i=i+1
    end
  end
  self.imagecache[name] = image
end

function Graphics:drawTile(x, y, t, c, flipped, upsideDown)
  local quad, name = self.quads[t], self.tileimage[t]
  if not (quad or name) then return end
  local image = self.imagecache[name]
  x, y = floor(x), floor(y)
  if c then self:setColor(c) end
  if flipped then x = x + 15 end
  if upsideDown then y = y + 15 end

  love.graphics.drawq(image, quad, x, y, 0, flipped and -1 or 1, upsideDown and -1 or 1)
  return self
end

function Graphics:rect(x, y, w, h)
  love.graphics.rectangle("fill", x, y, w, h)
  return self
end

function Graphics:start()
  love.graphics.scale(Graphics.scale, Graphics.scale)
  love.graphics.setLine(1, "rough")
  self:setColor(self.Color.White)
  return self
end

function Graphics:stop()
  love.graphics.setScissor()
  return self
end

return Graphics

