local floor, ceil, max, min = math.floor, math.ceil, math.max, math.min

--------------------------------------------------------------------------------
-- Map class

local Map = class {
  type = "Map",
}

local ENTITY_LAYER = "Entity Layer"
local TILE_LAYER = "Tile Layer"

function Map:init(name)
  self.data = util.loadLuaFile(name)
  for _, set in ipairs(self.data.tilesets) do
    Graphics:loadTileset(set.name, set.image:match("(gfx/.+)$"), set.firstgid)
  end
  self.layers = {}
  for _, layer in ipairs(self.data.layers) do
    if layer.name == ENTITY_LAYER then
      self.layers[ENTITY_LAYER] = layer
    elseif layer.name == TILE_LAYER then
      self.layers[TILE_LAYER] = layer
    end
  end
  self.width = self.data.width
  self.height = self.data.height
end

function Map:draw(wx, wy, mx, my, mw, mh)
  local left, right = floor(mx), ceil(mx+mw)
  local top, bottom = floor(my), ceil(my+mh)

  local tile_layer = self.layers[TILE_LAYER]
  local ent_layer = self.layers[ENTITY_LAYER]

  for yy = top, bottom do
    for xx = left, right do
      if Main.debug_window_enabled then
        Graphics:drawTile(xx*16-wx, yy*16, ent_layer.data[1 + yy*ent_layer.width + xx])
      end
    end
  end
end

function Map:getEntity(x, y)
  local layer = self.layers[ENTITY_LAYER]
  if x < 0 or y < 0 or x>= layer.width or y >= layer.height then return 0 end
  return layer.data[1+y*layer.width+x]
end

do
  local HINT, NUDGE = 0.001, 0.01
  local wall_mask = {
    [0] = { top=false,  bottom=false, left=false, right=false },
    [252] = { top=true,  bottom=false, left=false, right=false },
    [253] = { top=false, bottom=true,  left=false, right=false },
    [254] = { top=false, bottom=false, left=true,  right=false },
    [255] = { top=false, bottom=false, left=false, right=true  },
    [256] = { top=true,  bottom=true,  left=true,  right=true  },
  }

  local edgeMask = function(self, x, y)
    local mask = wall_mask[self:getEntity(x, y)] or wall_mask[0]
    return mask.top, mask.bottom, mask.left, mask.right
  end

  local between = function(x, y, z)
    return (x <= y) and (y <= z)
  end

  local colinear = function(a1, a2, b1, b2)
    return not ((a2<b1) or (a1>b2))
  end

  local function checkRun(self, dx, dy, ex, ey, aw, ah)
    local horiz_collision, vert_collision
    local ax, ay, bx, by
    local l, r = floor(ex/16)-1, floor((ex+aw)/16)+1
    local t, b = floor(ey/16)-1, floor((ey+ah)/16)+1
    for y = t, b do
      for x = l, r do
        local top_edge, bottom_edge, left_edge, right_edge = edgeMask(self, x, y)
        if top_edge or bottom_edge or left_edge or right_edge then
          ax, ay = ex + (dx or 0), ey + (dy or 0)
          bx, by = x*16, y*16
          if dx then
            if (dx > 0) and left_edge then
              if between(ex+aw, bx, ax+aw) and colinear(ay, ay+ah, by, by+16) then
                dx = max(-0.0, (dx - (ax+aw - bx+NUDGE)))
                horiz_collision = "right"
              end
            elseif (dx < 0) and right_edge then
              if between(ax, bx+16, ex) and colinear(ay, ay+ah, by, by+16) then
                dx = min(0.0, (dx + (bx+16+NUDGE - ax)))
                horiz_collision = "left"
              end
            end
          end
          if dy then
            if (dy > 0) and top_edge then
              if between(ey+ah, by, ay+ah) and colinear(ax, ax+aw, bx, bx+16) then
                dy = max(-0.0, (dy - (ay+ah - by+NUDGE)))
                vert_collision = "bottom"
              end
            elseif (dy < 0) and bottom_edge then
              if between(ay, by+16, ey) and colinear(ax, ax+aw, bx, bx+16) then
                dy = min(0.0, (dy + (by+16+NUDGE - ay)))
                vert_collision = "top"
              end
            end
          end
        end
      end
    end
    return dx, dy, horiz_collision, vert_collision
  end

  function Map:resolveEntityCollisions(dx, dy, ex, ey, ew, eh)
    local horiz, vert, _
    dx, _, horiz   = checkRun(self, dx, nil, ex, ey, ew, eh)
    _, dy, _, vert = checkRun(self, nil, dy, ex+dx, ey, ew, eh)
    return dx, dy, horiz, vert
  end
end

return Map

